#include <iostream>
#include <fstream>
#include <string>
#include <regex>
#include <cstdlib>
#include <vector>
#include <opencv2/opencv.hpp>
#include <selfbalance.hpp>


using namespace std;
using namespace cv;


struct Shape
{
  Shape(int value_treeC=0, int typeC=0, float aC=0, float bC=0, float offset_xC=0, float offset_yC=0, float rotateC=0, float center_xC=0, float center_yC=0, int indexC=0):type(typeC), a(aC), b(bC), offset_x(offset_xC), offset_y(offset_yC),  value_tree(value_treeC), rotate(rotateC), center_x(center_xC), center_y(center_yC), index(indexC) {}


  void cal_r_theta(float t) {
    // Circle
    if (type==0) {
      r_theta = b;
    }
    // Elipse
    else if (type==1) {
      r_theta = (a*b)/sqrt(pow(b*cos(t),2)+pow(a*sin(t),2));
    }
    //Cassini oval
    else{
      r_theta = sqrt(2*pow(a,2)*cos(2*t)+2*sqrt((-pow(a,4)+pow(b,4)) + pow(a,4)*pow(cos(2*t),2)))/2;
    }
  }

  float cal_x(float t) {
    return r_theta*sin(t-rotate)+center_x+offset_x;
  }

  float cal_y(float t) {
    return r_theta*cos(t-rotate)+center_y+offset_y;
  }


  float a,b,offset_x,offset_y,rotate,r_theta,center_x,center_y;
  int type;
  int value_tree, index;
};



// Helper function for generating float number
float RandomFloat(float min, float max)
{
  assert(max > min);
  float random = ((float) rand()) / (float) RAND_MAX;
  float range = max - min;
  return (random*range) + min;
}


// Generate a random curve (circle, elipse or cassini oval)
Shape gen_curve(float center_x, float center_y, float sep, int index) {
  float b = 30;
  float a = RandomFloat(0,b);
  int type = rand() % 3;
  float rotate = RandomFloat(-M_PI, M_PI);
  float offset_x = RandomFloat(b, sep-b);
  float offset_y = RandomFloat(b, sep-b);
  Shape rand_shape(center_x+center_y*1000, type, a, b, offset_x, offset_y, rotate,center_x,center_y,index);
  return rand_shape;
}

// Visualize a given equation of a shape
Mat draw_curve(Shape rand_shape) {
  float start_point_t = 0;
  float end_point_t = 2*M_PI;
  vector<Point2f> curvePoints;
  for (float t = start_point_t; t <= end_point_t; t+=0.1){
    rand_shape.cal_r_theta(t);
    float x = rand_shape.cal_x(t);
    float y = rand_shape.cal_y(t);
    Point2f new_point = Point2f(x, y);
    curvePoints.push_back(new_point);
  }
  Mat curve(curvePoints, true);
  curve.convertTo(curve, CV_32S);
  return curve;
}

// Generate a few random shapes
void gen_data(vector<Shape>& data, int im_size, int sep, avl_tree& tree) {
  int count = 0;
  for (int i=0; i<im_size; i+=sep){
    for (int j=0; j<im_size; j+=sep) {
      int cen_x = i;
      int cen_y = j;
      Shape current_shape = gen_curve(cen_x, cen_y, sep, count);
      data.push_back(current_shape);
      // Insert the element in the tree with an index (x+y*im_size)
      tree.r = tree.insert(tree.r, cen_x+cen_y*im_size, count);
      count++;
    }
  }
}


// Given a point and a shape, check if it is inside
void is_inside(Shape& data, float x, float y, Mat& img, int i) {
  float delta_x = x-(data.center_x+data.offset_x);
  float delta_y = y-(data.center_y+data.offset_y);
  float theta = atan2(delta_x, delta_y)+data.rotate;
  data.cal_r_theta(theta);
  bool inside = sqrt(pow(delta_x,2)+pow(delta_y,2))<data.r_theta;
  if (inside) {
    img.at<Vec3b>(y,x) = Vec3b(0,0,0);
  }
}


// Check whether a point is inside any of the shapes (naive)
void check_inside(vector<Shape>& data, float x, float y, Mat& img) {
  for (int i=0; i<data.size(); i++) {
    is_inside(data[i], x, y, img, i);
  }
}

// Check whether a point is inside any of the shapes (with a sorted tree)
void check_inside(vector<Shape>& data, int x, int y, Mat& img, avl_tree& tree, int im_size, int sep) {
  // DELETE BELOW LINE (int i = 0;) AND FILL IN
  int i = 0;
  is_inside(data[i], x, y, img, i);
}




int main() {
  avl_tree tree;
  int im_size = 1000;
  int sep = 100;
  Mat img(im_size, im_size, CV_8UC3, Scalar(255,255,255));
  vector<Shape> shapes;
  gen_data(shapes, im_size, sep, tree);
  // Balance the sorted tree
  tree.r = tree.balance(tree.r);
  for (int i=0; i<shapes.size(); i++){
    Mat circle_curve = draw_curve(shapes[i]);
    polylines(img, circle_curve, false, Scalar(0,0,0), 2, CV_AVX);
  }

  for (int i=0; i<im_size; i+=sep/40){
    for (int j=0; j<im_size; j+=sep/40) {
      // Search naively
      check_inside(shapes, i, j, img);
      // Search with a sorted tree
      // check_inside(shapes, i, j, img, tree, im_size, sep);
      }
  }


  imshow("image", img);
  waitKey();
  return 0;
};
